package de.uni_marburg.powersort.MSort;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.stream.IntStream;

public class IMPL_M_1 {

    private IMPL_M_1() {
    }

    /**
     * Sorts the given range, using the given workspace array slice
     * for temp storage when possible. This method is designed to be
     * invoked from public methods (in class Arrays) after performing
     * any necessary array bounds checks and expanding parameters into
     * the required forms.
     *
     * @param a        the array to be sorted
     * @param lo       the index of the first element, inclusive, to be sorted
     * @param hi       the index of the last element, exclusive, to be sorted
     * @param c        the comparator to use
     * @param work     a workspace array (slice)
     * @param workBase origin of usable space in work array
     * @param workLen  usable size of work array
     * @since 1.8
     */
    protected static int MERGE_COST = 0;

    public static void fillWithAscRunsHighToLow(Integer[] A, int[] runLengths, int runLenFactor) {
        //A has a fixed size, but it doesn't have any meaningful values
        int n = A.length;
        //This ensures that the sum of runLengths multiplied by runLenFactor equals the list size n. If not, an AssertionError is thrown.
        assert IntStream.of(runLengths).sum() * runLenFactor == n;

        //System.out.println("IntStream Of run length output: "+IntStream.of(runLengths).sum());
        //IntStream.of(runLengths).forEach(System.out::println);
        for (int i = 0; i < n; i++) {
            //putting i in set a, while a is always the last index of n
            A[i] = n - i;
        }


        //For each value l in the array runLengths, do the following
        // runLengths = {2, 3, 5}, the loop will run three times, with l taking values 2, 3, and 5 respectively.
        int startIndex = 0;
        for (int l : runLengths) {
            int L = l * runLenFactor;
            // Sort the subarray from startIndex to startIndex+L manually
            for (int i = startIndex; i < startIndex + L - 1; i++) {
                for (int j = startIndex; j < startIndex + L - 1 - (i - startIndex); j++) {
                    if (A[j] > A[j + 1]) {
                        // Swap elements
                        int temp = A[j];
                        A[j] = A[j + 1];
                        A[j + 1] = temp;
                    }
                }
            }
            startIndex += L;
        }
    }


    static <T> T[] merge(T[] run1, T[] run2, Comparator<? super T> c) {
        /**
         * We need this because:
         * In Java, you cannot create generic arrays directly (can't write new T[size])
         * The workaround is to create an Object array and cast it to T[]
         * This casting generates a warning because the compiler can't verify the type safety at runtime due to Java's type erasure
         * **/
        @SuppressWarnings("unchecked")
        T[] result = (T[]) new Object[run1.length + run2.length];
        int i = 0, j = 0, k = 0;

        while (i < run1.length && j < run2.length) {
            //This comparison only works if the lists are sorted
            if (c.compare(run1[i], run2[j]) <= 0) {
                result[k++] = run1[i++];
            } else {
                result[k++] = run2[j++];
            }
        }
        // can be improved by finding out which one is empty and only add the other one
        // Copy remaining elements
        while (i < run1.length) {
            result[k++] = run1[i++];
        }
        while (j < run2.length) {
            result[k++] = run2[j++];
        }

        return result;
    }

    // New helper method to handle range reversal
    private static <T> void reverseRange(T[] a, int start, int end) {
        // end is exclusive, so we need to use end-1 as the right boundary
        int left = start;
        int right = end - 1;
        while (left < right) {
            T temp = a[left];
            a[left] = a[right];
            a[right] = temp;
            left++;
            right--;
        }
    }

    static <T> void mergeInplace(T[] a, int i, int m, int j, Comparator<? super T> c) {
        System.out.printf("Merge(%d, %d, %d)%n", i, m, j);
        MERGE_COST += j - i;
       // Create temporary arrays for merging
        @SuppressWarnings("unchecked")
        T[] left = (T[]) new Object[m - i];
        @SuppressWarnings("unchecked")
        T[] right = (T[]) new Object[j - m];

        // Copy data to temporary arrays
        System.arraycopy(a, i, left, 0, m - i);
        System.arraycopy(a, m, right, 0, j - m);

        // Merge back to original array
        // Merge the runs
        T[] merged = merge(left, right, c);

        //copy back to original array
        System.arraycopy(merged, 0, a, i,merged.length);
    }


    static <T> int extendRun(T [] a, int i, Comparator<? super T> c) {
        // if i was the element before end so just return the last element
        if (i == a.length - 1) {
            return i + 1;
        }
        //we're looking at the element next to a[i]
        int j = i + 1;
        if (c.compare(a[i],a[j]) <=0) {
            while (j < a.length && c.compare(a[j - 1], a[j]) <= 0) {
                j++;
            }
        } else {
            while (j < a.length && c.compare(a[j - 1],a[j]) > 0) {
                j++;
            }
            // Reverse the decreasing sequence
            reverseRange(a,i,j);
        }
        return j;
    }


    public static int power(int[] run1, int[] run2, int n) {
        int i1 = run1[0], n1 = run1[1];
        int i2 = run2[0], n2 = run2[1];

        assert i1 >= 0;
        assert i2 == i1 + n1;
        assert n1 >= 1 && n2 >= 1;
        assert i2 + n2 <= n;

        double a = ((i1 + n1 / 2.0d) / n);
        double b = ((i2 + n2 / 2.0d) / n);

        int l = 0;
        while (Math.floor(a * Math.pow(2, l)) == Math.floor(b * Math.pow(2, l))) {
            l++;
        }
        return l;
    }

    public static <T> void mergeTopmost2(T[] a, List<int[]> runs, Comparator<? super T> c) {
        assert runs.size() >= 2;

        int[] Y = runs.get(runs.size() - 2);
        int[] Z = runs.getLast();

        assert Z[0] == Y[0] + Y[1];

        mergeInplace(a, Y[0], Z[0], Z[0] + Z[1], c);

        runs.set(runs.size() - 2, new int[] {Y[0], Y[1] + Z[1], Y[2]});
        runs.removeLast();
    }

    public static <T> void powerSort(T[] a, Comparator<? super T> c) {
        int n = a.length;
        int i = 0;
        List<int[]> runs = new ArrayList<>();

        int j = extendRun(a, i, c);
        runs.add(new int[] {i, j - i, 0});
        i = j;

        while (i < n) {
            j = extendRun(a, i, c);
            int p = power(runs.getLast(), new int[] {i, j - i}, n);

            while (p <= runs.getLast()[2]) {
                mergeTopmost2(a, runs, c);
            }

            runs.add(new int[] {i, j - i, p});
            i = j;
        }

        while (runs.size() >= 2) {
            mergeTopmost2(a, runs, c);
        }
    }

    public static int extendRunIncreasingOnly(List<Integer> a, int i) {
        if (i == a.size() - 1) {
            return i + 1;
        }
        int j = i + 1;
        while (j < a.size() && a.get(j - 1) <= a.get(j)) {
            j++;
        }
        return j;
    }

    public static int powerFast(int[] run1, int[] run2, int n) {
        int i1 = run1[0], n1 = run1[1];
        int i2 = run2[0], n2 = run2[1];

        int a = 2 * i1 + n1;
        int b = a + n1 + n2;

        int l = 0;
        while (true) {
            l++;
            if (a >= n) {
                assert b >= a;
                a -= n;
                b -= n;
            } else if (b >= n) {
                break;
            }
            assert a < b && b < n;
            a <<= 1;
            b <<= 1;
        }
        return l;
    }
//
//    ,
//            new SortImpl("MSort") {
//        @Override
//        @SuppressWarnings("unchecked")
//        public void sort(Object[] a) {
//            // Create a list of type that matches powerSort's requirements
//            List<Comparable<Object>> list = new ArrayList<>();
//            for (Object obj : a) {
//                // Since we know the input will be Integer objects in the benchmark
//                list.add((Comparable<Object>) obj);
//            }
//
//            // Call powerSort with the proper type
//            IMPL_M_1.<Comparable<Object>>powerSort(list);
//
//            // Copy back to array
//            for (int i = 0; i < a.length; i++) {
//                a[i] = list.get(i);
//            }
//        }
//    }
}